from mwcleric.clients.cargo_client import CargoClient
from mwcleric.clients.site import Site
from .auth_credentials import AuthCredentials
from .wiki_client import WikiClient


class WikiGGClient(WikiClient):
    """
    Functions for connecting to and editing specifically wiki.gg wikis.
    """
    cargo_client: CargoClient = None
    client: Site = None
    wiki: str = None

    def __init__(self, wiki: str, client: Site = None,
                 credentials: AuthCredentials = None, lang: str = None,
                 **kwargs):
        """
        Create a site object.

        :param wiki: Name of a wiki
        :param lang: Optional. If the wiki has a language path in the URL, provide it here.
        :param client: Otpional. If this is provided, SessionManager will not be used.
        :param credentials: Optional. Provide if you want a logged-in session.
        """

        url = '{}.wiki.gg{}'.format(wiki, f"/{lang}" if lang is not None else '')
        path = '/' if lang is not None else '/wiki/'
        super().__init__(url=url, path=path, credentials=credentials, client=client, **kwargs)

        self.cargo_client = CargoClient(self.client)

    def relog(self):
        super().relog()
        self.cargo_client = CargoClient(self.client)



from mwcleric.wiki_client import WikiClient


class LogEntry(object):
    def __init__(self, log, site: WikiClient):
        self.site = site
        self.log_type = log['type']
        self.title = log['title']
        self.page = site.client.pages[self.title]
        self.logid = log['logid']
        self.comment = log['comment']
        self.user = log['user']



from wiki_client import WikiClient
from wikigg_client import WikiGGClient
from fandom_client import FandomClient
from auth_credentials import AuthCredentials
from template_modifier import TemplateModifierBase
from page_modifier import PageModifierBase


import json
import os

from .errors import InvalidUserFile


class AuthCredentials(object):
    file_pattern = 'wiki_account_{}.json'
    username = None
    password = None
    config_path = os.path.join(os.path.expanduser('~'), '.config', 'mwcleric')

    def __init__(self, username=None, password=None, user_file=None, start_over=False):
        """
        Stores username and password for future use with a WikiClient.
        Specify either user_file or both username and password.

        If using a file, files can be either located in the same directory as the code,
        or in the config directory of the user. If files in both locations exist,
        files in the former location will trump those in the latter.

        :param username: Username, this must include an @ if using a bot password
        :param password: Password, this is the actual value of the password, not the "name" of a "bot password"
        :param user_file: Either a file or a system variable as a nicknamed account to look for
        """
        if start_over:
            if user_file is None:
                raise ValueError('No user_file is defined.')
            start_over_confirm = input('Are you sure you want to overwrite your previous credentials? [Y]es or [N]o ')
            if start_over_confirm.lower()[0] == 'y':
                account_data = self.prompt_user_info()
                self.save_account_data(account_data, user_file)
            else:
                # Proceeding with existing login credentials.
                pass

        if username and password:
            self.username = username
            self.password = password
        elif user_file:
            # Environment Variables Method
            pw = os.getenv('WIKI_PASSWORD_{}'.format(user_file.upper()), None)
            usr = os.getenv('WIKI_USERNAME_{}'.format(user_file.upper()), None)
            if pw and usr:
                self.password = pw
                self.username = usr
                return

            # Files / User Input Method
            user_info = self.get_user_data_from_file(user_file, '')
            if not user_info:
                user_info = self.get_user_data_from_file(user_file, self.config_path)
            if not user_info:
                account_data = self.prompt_user_info()
                self.save_account_data(account_data, user_file)
                user_info = self.get_user_data_from_file(user_file, self.config_path)
            if not user_info:
                raise InvalidUserFile
            self.password = user_info['password']
            self.username = user_info['username']

    def get_user_data_from_file(self, user_file, base_path):
        account_file = os.path.join(base_path, self.file_pattern.format(user_file.lower()))
        legacy_user_file = os.path.join(base_path, 'username_{}.txt'.format(user_file.lower()))
        legacy_pw_file = os.path.join(base_path, 'password_{}.txt'.format(user_file.lower()))
        if os.path.exists(account_file):
            return self.read_user_info(account_file)
        elif os.path.exists(legacy_user_file) and os.path.exists(legacy_pw_file):
            self.update_legacy_user_info(legacy_user_file, legacy_pw_file, account_file)
            return self.read_user_info(account_file)
        else:
            return None

    @staticmethod
    def read_user_info(file):
        with open(file) as f:
            json_data = json.load(f)
        return json_data

    @staticmethod
    def update_legacy_user_info(legacy_user_file, legacy_pw_file, account_file):
        with open(legacy_user_file, 'r') as f:
            username = f.read().strip()
        with open(legacy_pw_file, 'r') as f:
            password = f.read().strip()
        account_data = {
            'username': username,
            'password': password
        }
        with open(account_file, 'w') as f:
            f.write(json.dumps(account_data, indent=4))

    def save_account_data(self, account_data, user_file):
        os.makedirs(self.config_path, exist_ok=True)
        with open(os.path.join(self.config_path,
                               self.file_pattern.format(user_file.lower())), 'w') as f:
            f.write(json.dumps(account_data, indent=4))

    @staticmethod
    def prompt_user_info():
        print(
            'We will prompt for 3 separate things: Username, bot pw name, bot token name. Whitespace will be stripped.')
        username = input('What is your USERNAME (not bot password yet)?')
        pw_name = input('What is your bot pw NAME (not token yet)?')
        pw_token = input('What is your bot pw token/secret?')
        password = '{}@{}'.format(pw_name.strip(), pw_token.strip())
        account_data = {
            'username': username.strip(),
            'password': password,
        }
        return account_data



class Namespace(object):
    def __init__(self, id_number: int = None, name: str = None, canonical_name: str = None, aliases: list = None):
        self.id = id_number
        self.name = name
        self.canonical_name = canonical_name
        self.aliases = aliases or []



import re
from copy import deepcopy
from typing import Optional, Union

from mwparserfromhell.nodes import Template

from .errors import TemplateModifierNotImplemented
from .page_modifier import PageModifierBase
from .wiki_client import WikiClient


class TemplateModifierBase(PageModifierBase):
    def __init__(self, site: WikiClient, template, page_list=None, title_list=None, limit=-1, summary=None,
                 quiet=False, lag=0, tags=None, skip_pages=None,
                 recursive=True,
                 startat_page=None,
                 namespace: Optional[Union[int, str]] = None,
                 **data):
        """

        :param site: WikiClient site
        :param template: The template to modify
        :param page_list: A default page_list parameter. Otherwise the template's used_in list will be used
        :param title_list: See page_list.
        :param limit: See PageModifier class.
        :param summary: See PageModifier class.
        :param quiet: See PageModifier class.
        :param lag: See PageModifier class.
        :param tags: See PageModifier class.
        :param skip_pages: See PageModifier class.
        :param recursive: See mwparserfromhell.wikitext.filter_templates method
        :param startat_page: See PageModifier class
        :param namespace: Do we filter the template's used_in list?
        :param data: Extra keywords to save to the class for use in the update_template method
        """
        self.template_name = template
        self.current_template = None
        self.recursive = recursive
        self.check_deletion_marks = False
        self.invoke_namespace = None
        if not title_list:
            page_list = page_list if page_list else site.pages_using(template, namespace=namespace)
        super().__init__(site, page_list=page_list, title_list=title_list, limit=limit, summary=summary,
                         quiet=quiet, lag=lag, tags=tags, skip_pages=skip_pages,
                         startat_page=startat_page, **data)

    def update_wikitext(self, wikitext):
        for template in wikitext.filter_templates(recursive=self.recursive):
            name = template.name
            # handle Scribunto modules
            # see https://github.com/earwig/mwparserfromhell/issues/287
            if name.startswith('#invoke:'):
                if self.invoke_namespace is None:
                    self.invoke_namespace = self.site.ns_name_to_namespace.get('Module', False)
                if self.invoke_namespace:
                    name = deepcopy(name)
                    # replace with localized namespace
                    name.replace('#invoke:', self.invoke_namespace.name + ':')
            if name.matches(self.template_name):
                self.current_template = template
                self.update_template(template)

    def update_template(self, template):
        raise TemplateModifierNotImplemented()

    def remove_from_page(self):
        """
        Marks the template to remove it completely from the page in a later layer of processing
        """
        # we are not using wikitext.remove(node) because we want to clean up newlines later
        self.current_template: Template
        params = []
        for param in self.current_template.params:
            params.append(param.name.strip())
        for param in params:
            self.current_template.remove(param)

        self.current_template.name = '@@@DELETE@@@'
        self.check_deletion_marks = True

    def postprocess_plaintext(self, text):
        if not self.check_deletion_marks:
            return text

        # case: middle of article, with extra whitespace on each side
        text = text.replace('\n\n{{@@@DELETE@@@}}\n\n', '\n\n')
        # case: middle of article
        text = text.replace('\n{{@@@DELETE@@@}}\n', '\n')
        # case: beginning of article
        text = re.sub('^' + re.escape('{{@@@DELETE@@@}}\n'), '', text)
        # case: middle of a line or whatever
        text = text.replace('{{@@@DELETE@@@}}', '')
        return text



from mwcleric.models.logs.log_entry import LogEntry
from mwcleric.wiki_client import WikiClient


class MoveLogEntry(LogEntry):
    def __init__(self, log, site: WikiClient):
        super().__init__(log, site)
        self.title = log['params']['target_title']
        self.page = site.client.pages[self.title]



from mwparserfromhell import parse
from time import sleep

from .wiki_client import WikiClient


class PageModifierBase(object):
    """
    Update pages on a wiki by using update_wikitext or update_plaintext.

    Available attributes you can change:

    * self.current_text (do this in self.update_plaintext()
    * self.current_wikitext (do this in self.update_wikitext()

    Available attributes you can use but not modify:

    * self.current_page, which is an mwclient Page, so you can access page.name, etc
    * self.site, a WikiClient object

    Specify a summary on initialization along with either a page_list or a title_list.

    * page_list is a list of Page objects (for example maybe site.client.categories)
    * title_list is a list of strings which will be turned into Page objects
    """
    current_page = None
    current_text = None
    current_wikitext = None
    prioritize_plaintext = False
    prioritize_wikitext = False

    def __init__(self, site: WikiClient, page_list=None, title_list=None, limit=-1, summary=None, startat_page=None,
                 tags=None, skip_pages=None,
                 quiet=False, lag=0, **data):
        """Create a PageModifier object, which can perform operations to edit the plaintext
        or wikitext of a page.

        :param site: a WikiClient to perform the edits on
        :param page_list: a list of Page objects to operate on, do not use with title_list
        :param title_list: a list of title strings to operate on, do not use with page_list
        :param limit: stop after scanning this many pages
        :param summary: edit summary
        :param startat_page: skip to this page
        :param quiet: don't print any console output (set to True for cron processes)
        :param lag: sleep this many seconds before saving
        :param data: Extra keywords to save to the class for use in the update_wikitext/update_plaintext methods
        """
        if title_list is not None:
            page_list = [site.client.pages[p] for p in title_list]
        self.summary = summary if summary else 'Bot Edit'
        self.site = site
        self.page_list = page_list
        self.limit = limit
        self.passed_startat = False if startat_page else True
        self.startat_page = startat_page
        self.skip_pages = skip_pages if skip_pages is not None else []
        self.lag = lag
        self.quiet = quiet
        self.tags = tags
        self.data = data

    def _print(self, s):
        """Print iff the quiet flag is not set to True"""
        if self.quiet:
            return
        print(s)

    def update_wikitext(self, wikitext):
        """This will be run iff update_wikitext isn't overridden in a subclass.
        
        Modify wikitext in place.
        """
        self.prioritize_plaintext = True

    def update_plaintext(self, text):
        """This will be run iff update_plaintext isn't overridden in a subclass.
        
        Modify text and return it.
        """
        self.prioritize_wikitext = True
        return text

    def postprocess_plaintext(self, text):
        """This method may not be supported forever, do not use it!!!

        It's only here because TemplateModifier requires it due to mwparserfromhell's node removal
        not completely removing newlines when removing nodes.
        """
        return text

    def run(self):
        lmt = 0
        for page in self.page_list:
            if lmt == self.limit:
                break
            if self.startat_page and page.name == self.startat_page:
                self.passed_startat = True
            if not self.passed_startat:
                self._print("Skipping page %s, before startat" % page.name)
                continue
            if page.name in self.skip_pages:
                self._print("Skipping page %s as requested" % page.name)
                continue
            lmt += 1
            self.current_text = page.text()
            self.current_page = page
            self.current_wikitext = parse(self.current_text)
            self.current_text = self.update_plaintext(self.current_text)
            self.update_wikitext(self.current_wikitext)
            newtext = str(self.current_wikitext)

            # TODO: If mwparserfromhell has better support for removing nodes from wikitext,
            # delete postprocess_plaintext method
            newtext = self.postprocess_plaintext(newtext)
            if newtext != self.current_page.text() and not self.prioritize_plaintext:
                self._print('Saving page %s...' % page.name)
                sleep(self.lag)
                self.site.save(page, newtext, summary=self.summary, tags=self.tags)
            elif self.current_text != self.current_page.text():
                self._print('Saving page %s...' % page.name)
                sleep(self.lag)
                self.site.save(page, self.current_text, summary=self.summary, tags=self.tags)
            else:
                self._print('Skipping page %s...' % page.name)
